// PIC32MZ_14.c - Using SQI to read and write from an external PSRAM
// by Aidan Mocke 2019-02-25

// PIC32MZ2048EFH144 Configuration Bit Settings
//
// Please note: My PSRAM's chip select is on Port RJ1!
//

// DEVCFG3
// USERID = No Setting
#pragma config FMIIEN = ON              // Ethernet RMII/MII Enable (MII Enabled)
#pragma config FETHIO = ON              // Ethernet I/O Pin Select (Default Ethernet I/O)
#pragma config PGL1WAY = OFF            // Permission Group Lock One Way Configuration (Allow multiple reconfigurations)
#pragma config PMDL1WAY = OFF           // Peripheral Module Disable Configuration (Allow multiple reconfigurations)
#pragma config IOL1WAY = OFF            // Peripheral Pin Select Configuration (Allow multiple reconfigurations)
#pragma config FUSBIDIO = OFF           // USB USBID Selection (Controlled by Port Function)

// DEVCFG2
#pragma config FPLLIDIV = DIV_1         // System PLL Input Divider (1x Divider)
#pragma config FPLLRNG = RANGE_5_10_MHZ // System PLL Input Range (5-10 MHz Input)
#pragma config FPLLICLK = PLL_FRC       // System PLL Input Clock Selection (POSC is input to the System PLL)
#pragma config FPLLMULT = MUL_50        // System PLL Multiplier (PLL Multiply by 50)
#pragma config FPLLODIV = DIV_2         // System PLL Output Clock Divider (2x Divider)
#pragma config UPLLFSEL = FREQ_24MHZ    // USB PLL Input Frequency Selection (USB PLL input is 24 MHz)

// DEVCFG1
#pragma config FNOSC = SPLL             // Oscillator Selection Bits (System PLL)
#pragma config DMTINTV = WIN_127_128    // DMT Count Window Interval (Window/Interval value is 127/128 counter value)
#pragma config FSOSCEN = OFF            // Secondary Oscillator Enable (Disable SOSC)
#pragma config IESO = OFF               // Internal/External Switch Over (Disabled)
#pragma config POSCMOD = OFF            // Primary Oscillator Configuration (Disabled)
#pragma config OSCIOFNC = OFF           // CLKO Output Signal Active on the OSCO Pin (Disabled)
#pragma config FCKSM = CSECME           // Clock Switching and Monitor Selection (Clock Switch Enabled, FSCM Enabled)
#pragma config WDTPS = PS1048576        // Watchdog Timer Postscaler (1:1048576)
#pragma config WDTSPGM = STOP           // Watchdog Timer Stop During Flash Programming (WDT stops during Flash programming)
#pragma config WINDIS = NORMAL          // Watchdog Timer Window Mode (Watchdog Timer is in non-Window mode)
#pragma config FWDTEN = OFF             // Watchdog Timer Enable (WDT Disabled)
#pragma config FWDTWINSZ = WINSZ_25     // Watchdog Timer Window Size (Window size is 25%)
#pragma config DMTCNT = DMT31           // Deadman Timer Count Selection (2^31 (2147483648))
#pragma config FDMTEN = OFF             // Deadman Timer Enable (Deadman Timer is disabled)

// DEVCFG0
#pragma config DEBUG = OFF              // Background Debugger Enable (Debugger is disabled)
#pragma config JTAGEN = OFF             // JTAG Enable (JTAG Disabled)
#pragma config ICESEL = ICS_PGx1        // ICE/ICD Comm Channel Select (Communicate on PGEC1/PGED1)
#pragma config TRCEN = OFF              // Trace Enable (Trace features in the CPU are disabled)
#pragma config BOOTISA = MIPS32         // Boot ISA Selection (Boot code and Exception code is MIPS32)
#pragma config FECCCON = OFF_UNLOCKED   // Dynamic Flash ECC Configuration (ECC and Dynamic ECC are disabled (ECCCON bits are writable))
#pragma config FSLEEP = OFF             // Flash Sleep Mode (Flash is powered down when the device is in Sleep mode)
#pragma config DBGPER = PG_ALL          // Debug Mode CPU Access Permission (Allow CPU access to all permission regions)
#pragma config SMCLR = MCLR_NORM        // Soft Master Clear Enable bit (MCLR pin generates a normal system Reset)
#pragma config SOSCGAIN = GAIN_2X       // Secondary Oscillator Gain Control bits (2x gain setting)
#pragma config SOSCBOOST = ON           // Secondary Oscillator Boost Kick Start Enable bit (Boost the kick start of the oscillator)
#pragma config POSCGAIN = GAIN_2X       // Primary Oscillator Gain Control bits (2x gain setting)
#pragma config POSCBOOST = ON           // Primary Oscillator Boost Kick Start Enable bit (Boost the kick start of the oscillator)
#pragma config EJTAGBEN = NORMAL        // EJTAG Boot (Normal EJTAG functionality)
#pragma config CP = OFF                 // Code Protect (Protection Disabled)


#include <xc.h>
#include <stdio.h>

#define SYS_FREQ 200000000              // Running at 200MHz

unsigned char *TXDATA = (unsigned char *)&SQI1TXDATA;	// Address to write to for 8-bit data
unsigned char *RXDATA = (unsigned char *)&SQI1RXDATA;	// Address to read from for 8-bit data

void delay_us(unsigned int us)
{
    // Convert microseconds us into how many clock ticks it will take
    us *= (SYS_FREQ / 1000000) / 2; // Core Timer updates every 2 ticks
                      
    _CP0_SET_COUNT(0); // Set Core Timer count to 0

    while (us > _CP0_GET_COUNT()); // Wait until Core Timer count reaches the number we calculated earlier
}

void delay_ms(int ms)
{
    delay_us(ms * 1000);
}

void set_performance_mode()
{   
	unsigned int cp0;
	
    // Unlock Sequence
    asm volatile("di"); // Disable all interrupts
    SYSKEY = 0xAA996655;
    SYSKEY = 0x556699AA;  

    // PB1DIV
    // Peripheral Bus 1 cannot be turned off, so there's no need to turn it on
    PB1DIVbits.PBDIV = 1; // Peripheral Bus 1 Clock Divisor Control (PBCLK1 is SYSCLK divided by 2)

    // PB2DIV
    PB2DIVbits.ON = 1; // Peripheral Bus 2 Output Clock Enable (Output clock is enabled)
    PB2DIVbits.PBDIV = 1; // Peripheral Bus 2 Clock Divisor Control (PBCLK2 is SYSCLK divided by 2)

    // PB3DIV
    PB3DIVbits.ON = 1; // Peripheral Bus 2 Output Clock Enable (Output clock is enabled)
    PB3DIVbits.PBDIV = 1; // Peripheral Bus 3 Clock Divisor Control (PBCLK3 is SYSCLK divided by 2)

    // PB4DIV
    PB4DIVbits.ON = 1; // Peripheral Bus 4 Output Clock Enable (Output clock is enabled)
    while (!PB4DIVbits.PBDIVRDY); // Wait until it is ready to write to
    PB4DIVbits.PBDIV = 0; // Peripheral Bus 4 Clock Divisor Control (PBCLK4 is SYSCLK divided by 1)

    // PB5DIV
    PB5DIVbits.ON = 1; // Peripheral Bus 5 Output Clock Enable (Output clock is enabled)
    PB5DIVbits.PBDIV = 1; // Peripheral Bus 5 Clock Divisor Control (PBCLK5 is SYSCLK divided by 2)

    // PB7DIV
    PB7DIVbits.ON = 1; // Peripheral Bus 7 Output Clock Enable (Output clock is enabled)
    PB7DIVbits.PBDIV = 0; // Peripheral Bus 7 Clock Divisor Control (PBCLK7 is SYSCLK divided by 1)

    // PB8DIV
    PB8DIVbits.ON = 1; // Peripheral Bus 8 Output Clock Enable (Output clock is enabled)
    PB8DIVbits.PBDIV = 1; // Peripheral Bus 8 Clock Divisor Control (PBCLK8 is SYSCLK divided by 2)

    // PRECON - Set up prefetch
    PRECONbits.PFMSECEN = 0; // Flash SEC Interrupt Enable (Do not generate an interrupt when the PFMSEC bit is set)
    PRECONbits.PREFEN = 0b11; // Predictive Prefetch Enable (Enable predictive prefetch for any address)
    PRECONbits.PFMWS = 0b010; // PFM Access Time Defined in Terms of SYSCLK Wait States (Two wait states)

    // Set up caching
    cp0 = _mfc0(16, 0);
    cp0 &= ~0x07;
    cp0 |= 0b011; // K0 = Cacheable, non-coherent, write-back, write allocate
    _mtc0(16, 0, cp0);  

    // Lock Sequence
    SYSKEY = 0x33333333;
    asm volatile("ei"); // Enable all interrupts
 }

void setup_ports()
{
    // Turn off all analog inputs, make everything digital
    ANSELA = 0;
    ANSELB = 0;
    ANSELC = 0;
    ANSELD = 0;
    ANSELE = 0;
    ANSELF = 0;
    ANSELG = 0;
    ANSELH = 0;
    ANSELJ = 0;
    
    // Make all ports outputs initially
    TRISA = 0;
    TRISB = 0;
    TRISC = 0;
    TRISD = 0;
    TRISE = 0;
    TRISF = 0;
    TRISG = 0;
    TRISH = 0;
    TRISJ = 0;        
    
    // Set all port values low initially
    LATA = 0;
    LATB = 0;
    LATC = 0;
    LATD = 0;
    LATE = 0;
    LATF = 0;
    LATG = 0;
    LATH = 0;
    LATJ = 0;
}

void SQI_init()
{
     // Disable trace outputs because SQI shares them
    CFGCONbits.TROEN = 0;
	
    // Set up Reference Clock 2
    if (!REFO2CONbits.ACTIVE)
    {
        REFO2CONbits.RODIV = 1;
        REFO2CONbits.ROSEL = 1;
        REFO2CONbits.ON = 1;
        while (REFO2CONbits.DIVSWEN);
        REFO2CONbits.OE = 1;
    }
    
	// Turn *off* clock division according to the errata
    SQI1CLKCONbits.CLKDIV = 0;
    SQI1CLKCONbits.EN = 1;
	// Wait until the SQI clock reports it is stable
    while (!SQI1CLKCONbits.STABLE);
   
   // Tell the SQI peripheral to reset
    SQI1CFGbits.RESET = 1;    
    SQI1CFGbits.CPOL = 0;
    SQI1CFGbits.CPHA = 0;
	// Set the mode to 1, which is PIO mode (where we control it directly)
    SQI1CFGbits.MODE = 1;
	// Enable burst mode, again as datasheet says
    SQI1CFGbits.BURSTEN = 1;
	// Enable the SQI peripheral
    SQI1CFGbits.SQIEN = 1;
	// Enable data lines SQID0, SQID1, SQID2 and SQID3
    SQI1CFGbits.DATAEN = 0b10;
    
	// Set up buffers to trigger as soon as 1 byte is present
    SQI1THR = 0x100;
    SQI1INTTHR = 0x100;    
}

void SRAM_select(char select)
{
    LATJbits.LATJ1 = select;
}

void SRAM_get_EID(unsigned int *buf)
{    
	// Pull CS low, select the PSRAM
    SRAM_select(0);
    
	// Sending 4 bytes, 0x9F "Get EID" command and 3 empty address bytes
    SQI1CMDTHR = 0x100;
	// Deassert chip select when done, using device 1, using single lane mode, transmit command, 4 bytes
    SQI1CON = 0x00510004;
    // Remember, this next line actually sends 0x00 0x00 0x00 0x9F!
    SQI1TXDATA = 0x9F;
    
	// Wait until the transmit buffer is empty i.e. the data has been fully transmitted
    while (SQI1STAT1bits.TXBUFFREE < 32);
    
    // Trigger on each byte received
	SQI1CMDTHR = 0x01; // Receive 8 bytes
	// Deassert chip select when done, using device 1, using single lane mode, receive command, 8 bytes
    SQI1CON = 0x00520008;

    // Wait until the lower 8 bits of SQI1STAT1 (which are received bytes count) = 8
	while ((SQI1STAT1 & 0xFF) != 0x08);
    
    // Read the first 4 bytes and store them in the array pointer
	*buf = SQI1RXDATA;
	// Move array pointer to next element
    buf++;
	// Read the final 4 bytes and store them in the array pointer
    *buf = SQI1RXDATA;

	// Set CS high again, deselect the PSRAM
    SRAM_select(1);
}

void SRAM_send_data_quad(unsigned char data)
{
    unsigned char *TXDATA;

    TXDATA = (unsigned char *)&SQI1TXDATA;

    // Set SQI TX command threshold to start transfer after 1 byte
    SQI1CMDTHR = 0x100; 
    // Deassert chip select when done, using device 1, using quad lane mode now, transmit command, 1 byte
    SQI1CON = 0x00590001;

    // Send an 8-bit value to the buffer
    *TXDATA = data;

    // Wait until the transmit buffer is empty i.e. the data has been fully transmitted
    while (SQI1STAT1bits.TXBUFFREE < 32);
}

// Send the Quad Write command 0x38 followed by a 24-bit address to the PSRAM chip
void SRAM_start_write_quad(int address)
{
	// We're about to have some endian fun
    unsigned int data;
    unsigned int endian[3];
    
	// Start transfer after 4 bytes are in the buffer
	SQI1CMDTHR = 0x00000400;

    // Deassert chip select when done, using device 1, using quad lane mode now, transmit command, 4 bytes
	SQI1CON = 0x00590004;
    
    // Now, if our address is 0x123456, we are going to have to switch that around to 0x563412 because the PIC32 is a little-endian device. Yay.  
	endian[0] = address >> 16;
    endian[1] = (address & 0x00FF00) >> 8;
    endian[2] = (address & 0xFF);
    
	address = (endian[2] << 16) | (endian[1] << 8) |(endian[0]);
    data = (address << 8) | 0x38;

    // The actual order of bytes *sent* will be 0x38, address[16:23], address[8:15], address[0:7].
	// I've done it in a bit of a round-about way to hopefully make it clearer.
	
    // Send the 4 bytes of data to the transfer queue
	SQI1TXDATA = data;

	// Wait until the transmit buffer is empty i.e. the data has been fully transmitted
    while (SQI1STAT1bits.TXBUFFREE < 32);
}

// Send the Quad Read command 0xEB followed by a 24-bit address to the PSRAM chip
void SRAM_start_read_quad(int address)
{
	// We're about to have some endian fun
    unsigned int data;
    unsigned int endian[3];
    
	// Start transfer after 4 bytes are in the buffer
	SQI1CMDTHR = 0x00000400;

    // Deassert chip select when done, using device 1, using quad lane mode now, transmit command, 4 bytes
	SQI1CON = 0x00590004;
        
    // Now, if our address is 0x123456, we are going to have to switch that around to 0x563412 because the PIC32 is a little-endian device. Yay.  
	endian[0] = address >> 16;
    endian[1] = (address & 0x00FF00) >> 8;
    endian[2] = (address & 0xFF);
    
	address = (endian[2] << 16) | (endian[1] << 8) |(endian[0]);
    data = (address << 8) | 0xEB;

    // The actual order of bytes *sent* will be 0xEB, address[16:23], address[8:15], address[0:7].
	// I've done it in a bit of a round-about way to hopefully make it clearer.

    SQI1TXDATA = data;

    while (SQI1STAT1bits.TXBUFFREE < 32);
}

unsigned char SRAM_read_byte_quad()
{
    unsigned char result;
    
    // Trigger operation after 1 byte is in the buffer
    SQI1CMDTHR = 1;
    
    // Deassert chip select when done, using device 1, using quad lane mode now, receive command, 1 byte
    SQI1CON = 0x005A0001;
    
    // Wait until we have 1 byte in the RX buffer
    while ((SQI1STAT1 & 0xFF) != 1);

    // Read the 8-bit value from SQI1RXDATA indirectly
    result = *RXDATA;
    
    return result;
}


void SRAM_go_SQI()
{
	// We want to write an 8-bit value to the SQI transmit buffer, so set that up
    unsigned char *TXDATA;
    TXDATA = (unsigned char *)&SQI1TXDATA;

	// Pull CS low, select the PSRAM
    SRAM_select(0);

    // Tell the SQI peripheral to start transfer once 4 bytes are in the TX buffer
    SQI1CMDTHR = 0x100;
    
	// Deassert chip select when done, using device 1, using single lane mode, transmit command, 1 byte
    SQI1CON = 0x00510001;
    
	// Write the 8-bit value 0x35 to the transmit buffer
    *TXDATA = 0x35;
    
	// Wait until the transmit buffer is empty i.e. the data has been fully transmitted
    while (SQI1STAT1bits.TXBUFFREE < 32);
    
	// Set CS high again, deselect the PSRAM
    SRAM_select(1);
}

void SRAM_read(unsigned int address, unsigned char *dest, int num_bytes, unsigned int *num_read)
{
    int cnt;
    static unsigned char data;
    
	// Pull CS low, select the PSRAM
    SRAM_select(0);
    
    // Tell the PSRAM we are wanting to perform a Quad Read from address
    SRAM_start_read_quad(address);
    
    // Read 3 bytes of data (wait cycles))
    data = SRAM_read_byte_quad();
    data = SRAM_read_byte_quad();
    data = SRAM_read_byte_quad();
    
    for (cnt = 0; cnt < num_bytes; cnt++)
        *dest++ = SRAM_read_byte_quad();

	// Set CS high again, deselect the PSRAM
    SRAM_select(1);
    
    *num_read = num_bytes;
}

void SRAM_write_buffer(int address, unsigned char *buffer, int num_bytes)
{
	int cnt;
	
    // Select the SRAM chip by setting Chip Select to 0
	SRAM_select(0);
	
    // Tell the chip we want to start writing to the address contained in [address]
    SRAM_start_write_quad(address);
	
	// Write the data one byte at a time
    for (cnt = 0; cnt < num_bytes; cnt++)
    {
        SRAM_send_data_quad(*buffer++);
    }
	
    // De-select the SRAM chip by setting Chip Select to 1
    SRAM_select(1);
}

int main()
{
    int cnt;
    unsigned char buffer[256];
    unsigned int temp[2];
    
    set_performance_mode();

    setup_ports();
    
    // De-select the SRAM chip by setting Chip Select to 1
	SRAM_select(1);
    
    // Setup SQI peripheral in single channel mode
    SQI_init();
    
    // Get EID data
    SRAM_get_EID(temp);
    
    // Fill the buffer with values 0 to 255
    for (cnt = 0; cnt < 256; cnt++)
        buffer[cnt] = cnt;
    
    // Switch SRAM to quad lane mode
    SRAM_go_SQI();
    
    // Write data in buffer to SRAM starting at address 0
    SRAM_write_buffer(0, buffer, 256);

    // Fill the buffer with 0, 1, 2, 3 repeating
    for (cnt = 0; cnt < 256; cnt++)
    {
        buffer[cnt] = cnt % 4;
    }
    
    // Write data in buffer to SRAM starting at address 256
    SRAM_write_buffer(256, buffer, 256);
	    
    // Read 256 bytes of data into buffer from SRAM starting from address 0
    SRAM_read(0, buffer, 256, &temp[0]);

    // Read 256 bytes of data into buffer from SRAM starting from address 256
    SRAM_read(256, buffer, 256, &temp[0]);
}